/*
** Module   : FontManager.java
** Abstract : client-side font management class
**
** Copyright (c) 2011-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20111115 Created initial version
** 002 CA  20140711 Enhanced font support.
** 003 CA  20140805 Fixed warnings related to widget generics.
** 003 CA  20140805 Fixed warnings related to widget generic's.
** 004 MAG 20140910 Added unload notification.
** 005 HC  20140911 GUI coordinate support - data types refactoring.
** 006 HC  20141021 Implemented window move/position support.
** 007 CA  20141123 Allow the GUI client to be reinitialized.
** 008 CA  20141125 USE-3D-SIZE property is set by the ThinClient, on client initialization (to 
**                  allow configuration via the client's bootstrap config).
** 009 CA  20141220 Fixed constant for DEFAULT-FIXED-FONT; empty text length in GUI is always 0.
** 010 CA  20150317 Added tracking of fixed-size fonts and font's max width.
** 011 CA  20150406 Cache the text metrics on client-side, when they are resolved first time.
** 012 GES 20150428 Rewrite to eliminate Swing Font class dependencies. Now it is the FontDetails
**                  instance which is always used by common code.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;
import java.util.List;
import java.util.logging.*;

import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.driver.ScreenDriver;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * Font configuration manager, for a specific client session.
 */
public class FontManager
{
   /**
    * Reference for the undefined font.  This instructs the runtime to compute the font from the
    * parent's specification (i.e. from the frame/window/etc).
    */
   public static final int UNDEFINED_FONT = -3;

   /**
    * Reference for the default font.  0 is a valid entry in the font-table, so it can't be used.
    */
   public static final int DEFAULT_FONT = -1;

   /**
    * Reference for the default fixed font.  0 is a valid entry in the font-table, so it can't be 
    * used.
    */
   public static final int DEFAULT_FIXED_FONT = -2;

   /** The base for the system fonts. */
   public static final int SYSTEM_FONT_BASE = -1000;
   
   /** Reference to the window title font, a system font. */
   public static final int WINDOW_TITLE_FONT = SYSTEM_FONT_BASE + 0;
   
   /** logger */
   private static final Logger LOG = LogHelper.getLogger(GenericFrame.class.getName());
   
   /** Context local instance of this class. */
   private static final ContextLocal<WorkArea> local = new ContextLocal<WorkArea>()
   {
      protected WorkArea initialValue()
      { 
         // TODO: remove ThinClient dependency
         ThinClient tc = ThinClient.getInstance();
         return new WorkArea(tc.getServer());
      };
   };
   
   /**
    * Get the height of the given font.
    * 
    * @param    window
    *           The window from which the font-table is used.
    * @param    font
    *           Font index.
    * @param    driver
    *           The reference screen driver.
    *          
    * @return   Height of the font.
    */
   public static int getFontHeight(TitledWindow<?> window, int font, ScreenDriver<?> driver)
   {
      if (driver.isChui())
      {
         return 1;
      }
      
      WorkArea wa = local.get();
      
      String key = getFontKey(window, font, driver);
      FontDetails<?> details = wa.fontCache.get(key);
      int height = details.legacyHeight;
      
      if (height == -1)
      {
         // log warning that a font with no legacy height is used.
         if (LOG.isLoggable(Level.WARNING))
         {
            final String msg = "Font [%d] with key [%s] has no legacy height.";
            LOG.log(Level.WARNING, String.format(msg, font, key));
         }

         // let the driver measure the font
         FontDetails<?> fd = wa.fontCache.get(key);
         
         height = ((GuiDriver)driver).getFontHeight(fd);
      }
      
      return height;
   }
   
   /**
    * Check if the system font from the given font table entry is a fixed-size font.
    * 
    * @param    window
    *           The window for which the font-table is retrieved.
    * @param    font
    *           The font number in the window's font table.
    * @param    driver
    *           The active screen driver.
    *           
    * @return   <code>true</code> if the associated font is fixed-size.
    */
   public static boolean isFixedSize(TitledWindow<?> window, int font, ScreenDriver<?> driver)
   {
      if (driver.isChui())
      {
         return true;
      }
      
      WorkArea wa = local.get();
      
      String key = getFontKey(window, font, driver);
      FontDetails<?> details = wa.fontCache.get(key);
      
      return details.fixedSize;
   }

   /**
    * Get the width of the given font.
    * 
    * @param    window
    *           The window from which the font-table is used.
    * @param    font
    *           Font index.
    * @param    driver
    *           The reference screen driver.
    *          
    * @return   Width of the font.
    */
   public static int getFontWidth(TitledWindow<?> window, int font, ScreenDriver<?> driver)
   {
      if (driver.isChui())
      {
         return 1;
      }
      
      WorkArea wa = local.get();
      
      String key = getFontKey(window, font, driver);
      FontDetails<?> details = wa.fontCache.get(key);
      int width = details.legacyWidth;
      
      if (width == -1)
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            final String msg = "Font [%d] with key [%s] has no legacy width.";
            LOG.log(Level.WARNING, String.format(msg, font, key));
         }

         // let the driver measure the font
         FontDetails<?> fd = wa.fontCache.get(key);
         
         width = ((GuiDriver)driver).getFontWidth(fd);
      }
      
      return width;
   }
   
   /**
    * Get the maximum font width for the given font.
    * 
    * @param    window
    *           The window for which the font-table is retrieved.
    * @param    font
    *           The font number in the window's font table.
    * @param    driver
    *           The active screen driver.
    * 
    * @return   The font's maximum width.
    */
   public static int getMaxFontWidth(TitledWindow<?> window, int font, ScreenDriver<?> driver)
   {
      if (driver.isChui())
      {
         return 1;
      }
      
      WorkArea wa = local.get();
      
      String key = getFontKey(window, font, driver);
      FontDetails<?> details = wa.fontCache.get(key);
      int width = details.legacyMaxWidth;
      
      if (width == -1)
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            final String msg = "Font [%d] with key [%s] has no legacy max width.";
            LOG.log(Level.WARNING, String.format(msg, font, key));
         }
         
         // let the driver measure the font
         width = ((GuiDriver)driver).getFontMaxWidth(details);
      }
      
      return width;
   }

   /**
    * Resolve a font instance associated with the specified font entry in the window's 
    * font table.
    * 
    * @param    window
    *           The window from which the font-table is used.
    * @param    font
    *           The entry in the font-table.
    * @param    driver
    *           The reference screen driver.
    *
    * @return   See above.
    */
   public static FontDetails<?> resolveFont(Window<?> window, int font, ScreenDriver<?> driver)
   {
      if (driver.isChui())
      {
         return null;
      }
      
      WorkArea wa = local.get();
      
      String key = getFontKey(window, font, driver);
      
      FontDetails<?> details = wa.fontCache.get(key);
      
      return details;
   }

   /**
    * Get the width of the specified text, in pixels, using the specified font.  
    * 
    * @param    text
    *           The text.
    * @param    window
    *           The window from which the font-table is used.
    * @param    font
    *           The entry in the font-table.
    * @param    driver
    *           The reference screen driver.
    *           
    * @return   See above.
    */
   public static int getTextWidth(String          text,
                                  Window<?>       window,
                                  int             font,
                                  ScreenDriver<?> driver)
   {
      if (driver.isChui() || text.length() == 0)
      {
         return text.length();
      }
      
      int[] metrics = getTextMetrics(text, window, font, driver);

      return metrics[0];
   }

   /**
    * Get the height of the specified text, in pixels, using the specified font.  
    * 
    * @param    text
    *           The text.
    * @param    window
    *           The window from which the font-table is used.
    * @param    font
    *           The entry in the font-table.
    * @param    driver
    *           The reference screen driver.
    *           
    * @return   See above.
    */
   public static int getTextHeight(String          text,
                                   Window<?>       window,
                                   int             font,
                                   ScreenDriver<?> driver)
   {
      if (driver.isChui())
      {
         return 1;
      }
      
      int[] metrics = getTextMetrics(text, window, font, driver);
      
      return metrics[1];
   }
   
   /**
    * Get the PIXELS-PER-COLUMN value for this client session.
    *
    * @param    driver
    *           The reference screen driver.
    * 
    * @return   See above.
    */
   public static int getPixelsPerColumn(ScreenDriver<?> driver)
   {
      WorkArea wa = locate();
      
      if (driver.isChui())
      {
         return CoordinatesConversionParams.CHUI_PIXELS_PER_COLUMN;
      }

      return wa.defaultFont.legacyWidth;
   }
   
   /**
    * Get the PIXELS-PER-ROW value for this client session.
    * 
    * @param    driver
    *           The reference screen driver.
    *           
    * @return   See above.
    */
   public static int getPixelsPerRow(ScreenDriver<?> driver)
   {
      WorkArea wa = locate();

      if (driver.isChui())
      {
         return CoordinatesConversionParams.CHUI_PIXELS_PER_ROW;
      }

      int legacyHeight = wa.defaultFont.legacyHeight;
      
      return legacyHeight + (int) (legacyHeight / 2.0) + (wa.isThreeD ? 2 : 0);
   }

   /**
    * Compute and register the window's font-table.  This may be from the window's enviroment
    * (if the window is using a custom environment), or from the default environment (i.e. the
    * in-directory font-table).
    * <p>
    * The same font-table may be shared by one or more windows.
    * 
    * @param    window
    *           The window.
    * @param    driver
    *           The reference screen driver.
    */
   public static void registerFontTable(Window<?> window, ScreenDriver<?> driver)
   {
      WorkArea wa = locate();
      
      if (driver.isChui())
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.warning("The font-table can be resolved only in GUI mode.");
         }

         // no-op for chui
         return;
      }

      // determine the environments's name
      String env = (wa.ed == null ? null : wa.ed.getEnvironmentName());
      if (env != null)
      {
         env = env.toLowerCase();
      }

      EnvironmentFontTable envft = wa.envFontTables.get(env);
      if (envft == null)
      {
         // read the current font-table from the current environment, if not already built
         String[] fonts = wa.server.readFontTable();
         
         envft = new EnvironmentFontTable(env, fonts);
         wa.envFontTables.put(env, envft);
      }

      // the font-table for a certain environment is common for all windows created with that
      // environment.
      wa.windowFontTables.put(window, envft);
   }
   
   /**
    * Deregister the font-table associated with this window.
    * 
    * @param    window
    *           The window.
    * @param    driver
    *           The reference screen driver.
    */
   public static void deregisterFontTable(Window<?> window, ScreenDriver<?> driver)
   {
      WorkArea wa = locate();
      
      if (driver.isChui())
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.warning("The font-table can be resolved only in GUI mode.");
         }

         // no-op for chui
         return;
      }

      wa.windowFontTables.remove(window);
   }

   /**
    * Set the number of entries in the font-table.
    * 
    * @param    window
    *           The window from which the font-table is used.
    * @param    numEntries
    *           The new size of the font-table.
    */
   public static void adjustFontTableSize(Window<?> window, int numEntries)
   {
      // TODO: what happens if the font-table is shrinked and there are widgets referring those
      // fonts?
  
      WorkArea wa = locate();
      EnvironmentFontTable  eft = wa.windowFontTables.get(window);

      while (eft.fonts.size() != numEntries)
      {
         if (numEntries < eft.fonts.size())
         {
            // shrink it
            eft.fonts.remove(eft.fonts.size() - 1);
         }
         else
         {
            // enlarge it with the default font
            eft.fonts.add(wa.defaultFont);
         }
      }
   }

   /**
    * Get the number of entries in the font-table.
    * 
    * @param    window
    *           The window from which the font-table is used.
    * 
    * @return   See above.
    */
   public static int fontTableSize(Window<?> window)
   {
      WorkArea wa = locate();
      EnvironmentFontTable  eft = wa.windowFontTables.get(window);
      
      return eft.fonts.size();
   }

   /**
    * Get a key describing the given font.
    * 
    * @param    name
    *           The font name.
    * @param    size
    *           The font point size.
    * @param    bold
    *           Flag indicating if the font style is bold.
    * @param    italic
    *           Flag indicating if the font style is italic.
    * @param    underline
    *           Flag indicating if the font style is underline.
    *           
    * @return   See above.
    */
   public static String getFontKey(String  name,
                                   int     size,
                                   boolean bold,
                                   boolean italic,
                                   boolean underline)
   {
      return String.format("%s,%d,%b,%b,%b", name.toLowerCase(), size, bold, italic, underline);
   }

   /**
    * Resolve the font key associated with the given font, in the current window's font-table.
    * 
    * @param    window
    *           The window from which the font-table is used.
    * @param    font
    *           The entry in the font-table.
    * @param    driver
    *           The reference screen driver.
    * 
    * @return   See above.
    */
   public static String getFontKey(TitledWindow<?> window, int font, ScreenDriver<?> driver)
   {
      if (driver.isChui())
      {
         return null;
      }
   
      WorkArea wa = local.get();
      EnvironmentFontTable ft = wa.windowFontTables.get(window);
      
      FontDetails<?> fd = null;
      
      if (font == UNDEFINED_FONT)
      {
         // the font should have been computed before this call.
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.severe("Undefined font is not allowed - the caller needs to specify a resolved " +
                       "font.");
         }
         
         font = DEFAULT_FONT;
      }
      
      if (font == DEFAULT_FONT)
      {
         fd = wa.defaultFont;
      }
      else if (font == DEFAULT_FIXED_FONT)
      {
         fd = wa.defaultFixedFont;
      }
      else if (font <= SYSTEM_FONT_BASE)
      {
         fd = wa.systemFonts[SYSTEM_FONT_BASE - font];
      }
      else if (font < 0 || font >= ft.fonts.size())
      {
         if (LOG.isLoggable(Level.FINE))
         {
            int sz = ft.fonts.size() - 1;
            
            LOG.fine(String.format("Font %d is not in the font-table (max-size is %d).", fd, sz));
         }

         // if the font is not referenced in the font-table, then the DefaultFont is used.
         fd = wa.defaultFont;
      }
      else
      {
         fd = ft.fonts.get(font);
      }
      
      return fd.key;
   }

   /**
    * Unload environment. A notification from environment API whenever an 
    * <code>UNLOAD environment</code> statement is executed.
    * 
    * @param   envName
    *          Environment name.
    */
   public static void unload(String envName)
   {
      WorkArea wa = local.get(false);
      if (wa != null && envName != null)
      {
         wa.envFontTables.remove(envName);         
      }
   }

   /**
    * Initialize the font manager's helpers. This is called every time the ThinClient is
    * initialized/re-initialized.
    * 
    * @param    server
    *           The server exports.
    */
   public static void init(ServerExports server)
   {
      WorkArea wa = local.get(false);
      
      if (wa != null)
      {
         wa.server = server;
      }
   }

   /**
    * Get the state of the <code>use-3d-size</code> flag.
    * 
    * @return   See above.
    */
   public static boolean isThreeD()
   {
      WorkArea wa = locate();
      
      return wa.isThreeD;
   }
   
   /**
    * Set the state of the <code>use-3d-size</code> flag.
    * 
    * @param    isThreeD
    *           The new state of this flag.
    */
   public static void setThreeD(boolean isThreeD)
   {
      WorkArea wa = locate();
      wa.isThreeD = isThreeD;
   }

   /**
    * If the given font is undefined, default to either the {@link #DEFAULT_FONT} or 
    * {@value #DEFAULT_FIXED_FONT}, depending on whether a fixed font is required or not.
    * 
    * @param    font
    *           The font number.
    * @param    fixedSize
    *           Flag indicating if a fixed-size font is needed, if the font is 
    *           {@link #UNDEFINED_FONT undefined}.
    * 
    * @return   The received font, if an explicit font number is given, or a resolved font,
    *           otherwise.
    */
   public static int resolveFont(int font, boolean fixedSize)
   {
      boolean explicit = (font != FontManager.UNDEFINED_FONT);
      return (explicit ? font : (fixedSize ? FontManager.DEFAULT_FIXED_FONT 
                                           : FontManager.DEFAULT_FONT));
   }
   
   /**
    * Locate the context-local data of this class.
    * 
    * @return   A {@link WorkArea} instance.
    */
   private static WorkArea locate()
   {
      return local.get();
   }
   
   /**
    * Get the metrics for the specified text and font from the
    * {@link WorkArea#textMetrics cache}.
    * <p>
    * If not present, ask the server-side for the metrics.
    * 
    * @param    text
    *           The text.
    * @param    window
    *           The window from which the font-table is used.
    * @param    font
    *           The entry in the font-table.
    * @param    driver
    *           The reference screen driver.
    * 
    * @return   See above.
    */
   private static int[] getTextMetrics(String          text,
                                       Window<?>       window,
                                       int             font,
                                       ScreenDriver<?> driver)
   {
      WorkArea wa = local.get();
   
      String key = getFontKey(window, font, driver);
      String tkey = key + "#" + text;
      
      int[] metrics = wa.textMetrics.get(tkey);
      
      if (metrics != null)
      {
         // found from cache, fast exit
         return metrics;
      }

      // check server-side
      metrics = wa.server.getLegacyTextMetrics(text, key);
      
      if (metrics == null)
      {
         // let the driver measure the string
         FontDetails<?> fd = wa.fontCache.get(key);
         
         int width = ((GuiDriver) driver).getTextWidth(text, fd);
         int height = ((GuiDriver) driver).getTextHeight(text, fd);
         
         metrics = new int[] { width, height };
      }
      
      // cache it
      wa.textMetrics.put(tkey, metrics);
      
      return metrics;
   }

   /**
    * Context-local data for this font manager.
    */
   private static class WorkArea
   {
      /** A font-key to font details map, containing all the fonts loaded by this client. */
      private final Map<String, FontDetails<?>> fontCache = new HashMap<String, FontDetails<?>>();
      
      /** A map of window to its in-use environment, as when it was created. */
      private final Map<Window<?>, EnvironmentFontTable> windowFontTables = new HashMap<>();
      
      /** A map of environment names to their associated font-table. */
      private final Map<String, EnvironmentFontTable> envFontTables = new HashMap<>();
      
      /** The default font for this session. */
      private final FontDetails<?> defaultFont;
      
      /** The default fixed font for this session. */
      private final FontDetails<?> defaultFixedFont;

      /** The system fonts for this session. */
      private final FontDetails<?>[] systemFonts;
      
      /** 
       * Flag identifying if this session was started in 3d.  Default value is <code>true</code>.
       */
      private boolean isThreeD = true;
      
      /** The server exports. */
      private ServerExports server;
      
      /** Access to the client's environment.  May be <code>null</code> if not on a Windows OS. */
      private EnvironmentDaemon ed;
      
      /** A cache of text metrics.  0-index is the width, 1-index is the height. */
      private final Map<String, int[]> textMetrics = new HashMap<String, int[]>();
      
      /**
       * Create a new work-area.
       * 
       * @param    server
       *           The server exports.
       */
      public WorkArea(ServerExports server)
      {
         if (OutputManager.getDriver().isChui())
         {
            this.server           = null;
            this.ed               = null;
            this.isThreeD         = false;
            this.defaultFont      = null;
            this.defaultFixedFont = null;
            this.systemFonts      = null;                           
         }
         else
         {
            // initialize only in GUI mode
            this.server   = server;
            
            // initialize the default fonts
            String[] defaults = server.readDefaultFonts();
   
            this.defaultFont      = createFont(defaults[0]);
            this.defaultFixedFont = createFont(defaults[1]);

            String[] sysFonts = server.readSystemFontTable();
            this.systemFonts = new FontDetails<?>[sysFonts.length];
            for (int i = 0; i < sysFonts.length; i++)
            {
               this.systemFonts[i] = createFont(sysFonts[i]);
            }
         }
      }
      
      /**
       * Create a new font, with the specified details.
       * <p>
       * It will access the server-side for details about the legacy metrics and/or custom 
       * definition.
       * 
       * @param    name
       *           The font name.
       * @param    size
       *           The font point size.
       * @param    bold
       *           Flag indicating if the font style is bold.
       * @param    italic
       *           Flag indicating if the font style is italic.
       * @param    underline
       *           Flag indicating if the font style is underline.
       *           
       * @return   A new {@link FontDetails} instance, with this font. 
       */
      public FontDetails<?> getFont(String  name, 
                                    int     size, 
                                    boolean bold, 
                                    boolean italic, 
                                    boolean underline)
      {
         if (size == 0)
         {
            if ("system".equals(name) || "fixedsys".equalsIgnoreCase(name))
            {
               size = 10;
               UnimplementedFeature.partial("Default size for System and FixedSys is set to 10!");
            }
            else
            {
               UnimplementedFeature.todo("The default font size is not computed yet - " +
                                         "font size requires to be specified!");
            }
         }
         
         String         key = getFontKey(name, size, bold, italic, underline);
         FontDetails<?> fd  = fontCache.get(key);
         
         if (fd == null)
         {
            int legacyWidth    = -1;
            int legacyMaxWidth = -1;
            int legacyHeight   = -1;
            
            byte[] fontDefinition = server.getCustomFontDefinition(name);
            int[]  legacyMetrics  = server.getLegacyFontMetrics(key);
            
            GuiDriver driver = (GuiDriver) OutputManager.getDriver();
            
            if (legacyMetrics != null)
            {
               legacyWidth    = legacyMetrics[0];
               legacyMaxWidth = legacyMetrics[1];
               legacyHeight   = legacyMetrics[2];
            }
            
            fd = driver.createFont(name,
                                   size,
                                   bold,
                                   italic,
                                   underline,
                                   fontDefinition,
                                   legacyWidth,
                                   legacyHeight);
            
            // finish configuration of the font details (by doing this here, the driver code
            // can be simpler, since all of these members can be determined in common code)
            
            fd.key = key;
            
            if (legacyMetrics == null)
            {
               // the driver methods for doing this require a font details instance and the
               // logic is complex enough that it doesn't make sense to duplicate it such that
               // the driver can efficiently do this itself so it is best to put this into
               // common code
               fd.legacyWidth    = driver.getFontWidth(fd);
               fd.legacyMaxWidth = driver.getFontMaxWidth(fd);
               fd.legacyHeight   = driver.getFontHeight(fd);
            }
            else
            {
               fd.legacyMaxWidth = legacyMaxWidth;
            }
            
            if (bold && italic)
            {
               fd.style = FontStyle.BOLD_ITALIC;
            }
            else if (bold)
            {
               fd.style = FontStyle.BOLD;
            }
            else if (italic)
            {
               fd.style = FontStyle.ITALIC;
            }
            else
            {
               fd.style = FontStyle.PLAIN;
            }
            
            fd.pointSize = size;
            fd.fixedSize = server.isLegacyFixedFont(name);
                                                                  
            fontCache.put(key, fd);
         }
         
         return fd;
      }
      
      /**
       * Create a new font, using the specified legacy definition.
       * 
       * @param    font
       *           The font definition.
       *           
       * @return   A {@link FontDetails} structure.
       */
      public FontDetails<?> createFont(String font)
      {
         boolean isBold = false;
         boolean isItalic = false;
         boolean isUnderline = false; // TODO:
         float ptSize = 0; // TODO: what if size is missing? is zero good?
         String fontName = "";
         
         if (font.indexOf(",") >= 0)
         {
            String[] tokens = font.split(",");
            for (int i = 1; i < tokens.length; i++)
            {
               String tok = tokens[i].trim().toLowerCase();
               if (tok.startsWith("size"))
               {
                  tok = tok.substring("size".length()).trim();
                  while (!Character.isDigit(tok.charAt(0)))
                  {
                     tok = tok.substring(1);
                  }
                  
                  ptSize = Float.parseFloat(tok);
               }
               else if (tok.startsWith("bold"))
               {
                  isBold = true;
               }
               else if (tok.startsWith("italic"))
               {
                  isItalic = true;
               }
            }
            
            fontName = tokens[0].trim();
         }
         else
         {
            fontName = font;                        
         }
   
         FontDetails<?> f = getFont(fontName, Math.round(ptSize), isBold, isItalic, isUnderline);
         
         return f;
      }   
   }
   
   /**
    * The font-table associated with a certain environment.
    */
   private static class EnvironmentFontTable
   {
      /** The enviroment's name. */
      @SuppressWarnings("unused")
      private final String envName;
      
      /** The enviroment's font-table. */
      private final List<FontDetails<?>> fonts;
      
      /**
       * Create a new font-table, for the specified environment.
       * 
       * @param    name
       *           The environment's name.
       * @param    fontTable
       *           The fonts part of this font-table.
       */
      public EnvironmentFontTable(String name, String[] fontTable)
      {
         this.envName = name;
         this.fonts = new ArrayList<>(fontTable.length);
         WorkArea wa = locate();
        
         for (int i = 0; i < fontTable.length; i++)
         {
            FontDetails<?> font = wa.createFont(fontTable[i]);
            this.fonts.add(font);
         }
      }
   }
}
