/*
** Module   : ChuiWebDriver.java
** Abstract : WebSocket-based ChUI driver
**
** Copyright (c) 2013-2015 Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 MAG 20131112 Created initial version.
** 002 MAG 20131129 Import server KeyStore.
** 003 MAG 20140109 Add Dojo modules handler and resource handler.
** 004 MAG 20140131 Add handler for /index.html.
**                  Add quit method.
** 005 MAG 20140212 Add more javadoc for method startupServer. 
** 006 CA  20140206 The jetty threads don't have context support, so any directory configuration
**                  must be read from the bootstrap config (i.e. socket timeout).
** 007 MAG 20140217 Add shutdown method.
** 008 MAG 20140303 Get authentication token from server key store. 
** 009 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
** 010 GES 20150312 Moved most of the embedded web server code to a common/shared class. 
*/

package com.goldencode.p2j.ui.client.chui.driver.web;

import java.util.logging.*;

import org.eclipse.jetty.server.handler.*;
import org.eclipse.jetty.util.resource.*;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.ui.client.chui.driver.swing.*;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.driver.web.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.web.*;
import com.goldencode.terminal.*;

/**
 * ChUI driver using WebSockets to communicate with a very thin javascript client hosted in
 * a browser. This class creates and uses an embedded web server to accomplish this.
 */
public class ChuiWebDriver
extends AbstractChuiDriver
implements EmbeddedWebServer
{
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(ChuiWebDriver.class.getName());

   /** Main page root target. */
   private static final String MAIN_TARGET = "/";

   /** Welcome page. */
   private static final String WELCOME_PAGE = "/index.html";
   
   /** Configuration. */
   private BootstrapConfig config;
   
   /** ChUI simulator. */
   private ChuiWebSimulator simulator;
   
   /** Delgated implementation for the embedded web server. */
   private EmbeddedWebServerImpl websrv;
   
   /**
    * Constructor.
    * 
    * @param   config
    *          The <code>BootstrapConfig</code> configuration.
    */
   public ChuiWebDriver(BootstrapConfig config)
   {
      super(new ChuiPrimitives(new ChuiWebSimulator(config)), "vt100");
      
      this.config    = config;
      this.simulator = (ChuiWebSimulator) ((ChuiPrimitives) direct).getEmulatedTerminalState();
   }
   
   /**
    * Initialize driver.
    */
   @Override
   public void init()
   {
      TerminalOptions options = (config == null) ? new TerminalOptions()
                                                 : convertToOptions(config);
      simulator.configure(options);
      
      // calculate our resource base
      Resource cpres = EmbeddedWebServerImpl.calcResourceBase(ChuiWebDriver.class, "res");
     
      // calculate our welcome page
      String welcomeUrl = getWelcomePageUrl(MAIN_TARGET, WELCOME_PAGE);
      
      // create our handlers
      ChuiWebPageHandler[] hdlrs = new ChuiWebPageHandler[2]; 
      
      hdlrs[0] = new ChuiWebPageHandler(config, MAIN_TARGET, welcomeUrl);         
      hdlrs[1] = new ChuiWebPageHandler(config, WELCOME_PAGE, welcomeUrl);
      
      WebClientProtocol websock = simulator.getWebSocket();
      
      // create the embedded server instance
      websrv = new EmbeddedWebServerImpl("ChuiWebDriver",
                                         new GenericWebSocketCreator<WebClientProtocol>(websock),
                                         cpres,
                                         hdlrs);
   }

   /**
    * Emit a sound beep on the terminal.
    * 
    */
   @Override
   public void beep()
   {
      if (simulator != null)
      {
         simulator.beep();
      }
   }
   
   /**
    * Clear screen.
    * <p>
    * This method should clear the screen and perform necessary flushing of the 
    * system-specific buffers if necessary. Users of this method assume that 
    * actual screen appearance as it is visible to the user is properly changed
    * upon exit from this method.
    */
   @Override
   public void clear()
   {
      simulator.clear();
   }

   /**
    * Get driver-specific child process factory.
    * 
    * @return   The reference to the factory.
    */
   @Override
   public ChildProcessFactory getChildProcessFactory()
   {
      return new ChildProcessFactory()
      {
         /**
          * Obtains a new instance of an interactive child process.
          *
          * @return   The new child process instance.
          */
         public InteractiveChildProcess getChildProcess()
         {
            return new VT100WebEmulator();
         }
      };
   }
   
   /**
    * Read single keystroke from keyboard. Returned code should be either proper
    * Progress 4GL/P2J key code or -1 if environment-specific input cannot be 
    * properly translated.
    * 
    * @return   The pressed key code or -1 if no keys were pressed since last call.
    */
   @Override
   public int readKey()
   {
      return simulator.readKey();
   }

   /**
    * Reset current screen mode.
    * <p>
    * Upon exit from this method P2J assumes that screen is in correct state
    * and can be used for interaction with the user.
    */
   @Override
   public void resetMode()
   {
      // NOP for now, what should this do?
   }

    /**
     * Resume normal driver operation. If implementation is terminal-based,
     * they terminal should be acquired again and driver should be ready to 
     * perform regular screen I/O operations.
     */
   @Override
   public void resume()
   {
   }

   /**
    * Suspend all screen I/O operations. If implementation is terminal-based, 
    * then terminal should be released to allow other application executed 
    * from P2J access terminal.
    */
   @Override
   public void suspend()
   {
   }
   
   /**
    * Set cursor status (ON/OFF).
    * <p>
    * The implementation of this method should show/hide actual screen 
    * cursor visible to user. Remaining code assumes that actual cursor status
    * visible to user is properly changed upon exit from this method.
    * 
    * @param   on
    *          <code>true</code> - show cursor, <code>false</code> - hide cursor.
    * @return  <code>true</code> if operation was successful.
    * 
    * @throws  IllegalStateException
    */
   @Override
   public boolean setCursorStatus(boolean on)
   throws IllegalStateException
   {
      return simulator.setCursorStatus(on);
   }
   
   /**
    * Perform driver shutdown.
    */
   public void shutdown()
   {
      // notify the remote client about application exit    
      simulator.quit();
      
      // since the application will exit shutdown the server
      shutdownServer();
   }
   
   /**
    * Startup the embedded server using the server's key store to build the web server SSL
    * context.
    * <p>
    * There are two page handlers.  One is for the root ("/") target and the other is for
    * the index ("/index.html") target.  Both load the same page index.html.
    * <p>
    * When the user is first redirected the URL is <code>https://server:port/</code> and the
    * target is "/". The index.html page is loaded for this target.
    * <p>
    * After redirection and the main page is loaded, the URL in the browser address bar will
    * become <code>https://server:port/index.html</code> and the corresponding target is now
    * "/index.html".
    * <p>
    * When the user reloads the main page within the browser environment the "/index.html"
    * target is sent to the web server and the same main page index.html is loaded. 
    * 
    * @param    keyStore
    *           The exported server KeyStore.
    * @param    config
    *           Client configuration.
    * @param    port
    *           An explicit port to start the web server or 0 to let the OS assign one.
    * @param    host
    *           An explicit host to start the web server or <code>null</code> to use the server's
    *           host.
    */
   @Override
   public String startupServer(ServerKeyStore keyStore, BootstrapConfig config, String host, int port)
   {
      return websrv.startupServer(keyStore, config, host, port);
   }
   
   /**
    * Shutdown the embedded web server.
    */
   @Override
   public void shutdownServer()
   {
      websrv.shutdownServer();
   }
   
   /**
    * Get the embedded web server's URI.
    * 
    * @return   A string representing the server's URI.
    */
   @Override
   public String getServerURI()
   {
      return websrv.getServerURI();
   }
   
   /**
    * Build the name of the main page as a resource locator string.
    * 
    * @param   pageContext
    *          Page context root.
    * @param   pageFile
    *          Main page HTML file.
    * 
    * @return  Resource name for the main page.
    */
   private String getWelcomePageUrl(String pageContext, String pageFile)
   {
      StringBuilder sb = new StringBuilder(pageContext);
      String packageName = ChuiWebPageHandler.class.getPackage().getName();
      return sb.append(packageName.replace('.', '/')).append(pageFile).toString();
   }
   
   /**
    * Web-specific implementation of the VT100 emulator child process.
    */
   private class VT100WebEmulator
   extends VT100EmulatorChildProcess
   {
      /**
       * Create a shell emulator and execute the given command.
       * 
       * @param    cmdline
       *           The command.
       *           
       * @return   A VT100 emulator executing the given command.
       */
      @Override
      public VT100Emulator createShellEmulator(String[] cmdline)
      {
         VT100WebRenderer vt100Renderer = new VT100WebRenderer(simulator.getScreenData(),
                                                               simulator.getRows(),
                                                               simulator.getColumns());
         
         VT100ProgressKeyboard vt100Keyboard = new VT100ProgressKeyboard();
         
         simulator.setVT100Keyboard(vt100Keyboard);
         
         // web clients are always an overlay
         int overlay = 1; 
         
         return new VT100Emulator(vt100Renderer, vt100Keyboard, cmdline, overlay);
      }
   }
   
   /**
    * Swing-specific implementation of the VT100 emulator renderer.
    */
   private class VT100WebRenderer
   extends VT100RendererImpl
   {
      /**
       * Constructor.
       * 
       * @param    data
       *           Video buffer.
       * @param    rows
       *           Number of rows on the screen.
       * @param    cols
       *           Number of columns on the screen.
       */
      public VT100WebRenderer(Cell[][]data, int rows, int cols)   
      {
         super(data, rows, cols);
      }
      
      /**
       * Ring the terminal bell.
       */
      @Override
      public void beep()
      {
         ChuiWebDriver.this.beep();
      }
      
      /**
       * Set cursor visibility.
       * 
       * @param    showCursor
       *           <code>true</code> to make the cursor visible, <code>false</code> to hide it.
       */
      @Override
      public void setCursorVisible(boolean showCursor)
      {
         simulator.setCursorStatus(showCursor);
         updateScreen();
      }
      
      /**
       * Update screen.
       */
      @Override
      protected void updateScreen()
      {
         simulator.triggerRepaint(true);
         simulator.setCursor(col, row);
      }
   }
}
