/*
** Module   : AbstractDriver.java
** Abstract : simple base class for common screen driver implementation
**
** Copyright (c) 2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 GES 20150401 Created initial version.
*/

package com.goldencode.p2j.ui.client.driver;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;

/**
 * Simple base class for common screen driver implementation features. This should not have
 * any features that are specific to ChUI or GUI. Instead, all features must be independent
 * of any UI modality.  In addition, no driver-specific implementation classes or features
 * should be present.  For example, no Swing-specific feature usage is allowed.
 */
public abstract class AbstractDriver<P extends OutputPrimitives, O extends OutputManager<P>>
implements ScreenDriver<O>
{
   /** This driver-specific widget factory */
   protected WidgetFactory<O> factory = null;
   
   /** Direct to screen output primitives. */
   protected P direct;
   
   /** UI modality flag: <code>true</code> for ChUI, <code>false</code> for GUI. */
   private boolean isChui;
   
   /** Batch mode flag. */
   private boolean batch = false;
   
   /** Background flag (signifies if the client is in the foreground session). */
   private boolean background = false;
   
   /** Terminal type name. */
   private String term;
   
   /**
    * Construct a new instance.
    *
    * @param    factory
    *           The widget factory to use for this driver.
    * @param    direct
    *           The output primitives.
    * @param    isChui
    *           <code>true</code> for ChUI, <code>false</code> for GUI.
    * @param    term
    *           Terminal type string.
    */
   public AbstractDriver(WidgetFactory<O> factory, P direct, boolean isChui, String term)
   {
      this.factory = factory;
      this.direct  = direct;
      this.isChui  = isChui;
      this.term    = term;
   }

   /**
    * Checks if the current driver is running in batch mode.
    * 
    * @return   Always <code>false</code>.
    * 
    */
   public boolean inBatchMode()
   {
      return batch;
   }

   /**
    * Checks if the current driver is running in background mode.
    * 
    * @return   Always <code>false</code>.
    * 
    */
   public boolean inBackgroundMode()
   {
      return background;
   }

   /**
    * Obtain system/implementation specific terminal type string.
    * <p>
    * At present this is used to implement 4GL TERMINAL function and do some 
    * terminal-specific workarounds. At present workarounds exist for "vt320" 
    * terminal type. If implementation requires no special workarounds, then 
    * it should avoid returning "vt320" as terminal type string.  
    * 
    * @return   The current terminal type string.
    */
   public String terminalType()
   {
      return term;
   }

   /**
    * Change terminal type to specified one. The terminal-based implementation
    * should try to change terminal type to specified one. It is assumed that 
    * passed terminal string is specific to particular driver implementation 
    * and recognized by it. 
    * 
    * @param    term
    *           The terminal type.
    */
   public void setTerminalType(String term)
   {
      this.term = term;
   }

   /**
    * Get driver-specific direct output primitives implementation.
    *
    * @return  driver-specific instance of direct screen output primitives.
    */
   public P getPrimitives()
   {
      return direct;
   }

   /**
    * Get driver-specific widget factory.
    * 
    * @return  reference to widget factory.
    */
   public WidgetFactory<O> getFactory()
   {
      return factory;
   }
   
   /**
    * Initialize driver.
    */
   @Override
   public void init()
   {
   } 

   /**
    * Get the default WINDOW-SYSTEM used by this driver. See {@link ClientParameters} for the list
    * of possible values.
    * 
    * @return   See above.
    */
   public String getWindowSystem()
   {
      return isChui ? ClientParameters.WINDOW_SYSTEM_TTY
                    : ClientParameters.WINDOW_SYSTEM_MS_WINDOWS;
   }
   
   /**
    * Get screen interface type.
    * 
    * @return  <code>true</code> is ChUI <code>false</code> is GUI
    */
   public boolean isChui()
   {
      return isChui;
   }

   /**
    * Perform driver shutdown.
    */
   public void shutdown()
   {
      // no-op
   }

   /**
    * Assigns the flag determining if the current driver is running in batch mode.
    * 
    * @param    batch
    *           The new state of the batch mode flag.
    * 
    */
   protected void setBatchMode(boolean batch)
   {
      this.batch = batch;
   }

   /**
    * Assigns the flag determining if the current driver is running in background mode.
    * 
    * @param    background
    *           The new state of the background mode flag.
    * 
    */
   protected void setBackgroundMode(boolean background)
   {
      this.background = background;
   }
}
