/*
** Module   : ContentPane.java
** Abstract : Swing-based printable content pane
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140805 Created initial version.  Extracted from code in previous ChuiSimulator class.
** 002 CA  20141029 Each simulator needs to receive the swing keyboard reader implementation in 
**                  use.  It's the callers responsibility to ensure that the same instance is used
**                  for all simulators, if the client is ran in a multi-window mode (i.e. GUI).
** 003 GES 20150416 Changed class name to more accurately reflect the purpose. This class is no
**                  longer used as a base class, but rather as a Swing component.
*/

package com.goldencode.p2j.ui.client.driver.swing;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.awt.print.*;
import java.util.*;

import javax.swing.*;

import com.goldencode.p2j.ui.client.chui.driver.*;

/**
 * Base class for all Swing-based simulators. 
 */
@SuppressWarnings("serial")
public class ContentPane
extends JComponent
implements Printable
{
   /** Containing frame (may be an applet or an application). */
   private Frame frame = null;
   
   /** The key reader associated with this simulator. */
   private final SwingKeyboardReader keyReader;
   
   /** Key processing for VT-100 emulator. */
   private VT100ProgressKeyboard keyboardVT100 = null;

   /**
    * Create a new instance with the container and frame given.  When this method is done, all 
    * listeners and key processing will be initialized.
    * <p>
    * For applications, the frame and the container are usually the same
    * object (of type <code>JFrame</code>). For applets, these are different
    * instances since a <code>JApplet</code> will be the container and the
    * frame will be a parent container of the <code>JApplet</code>.
    *
    * @param    frame
    *           The containing frame. Needed for key event processing
    *           and for the processing of modal dialog boxes.
    * @param    keyReader
    *           The {@link SwingKeyboardReader key reader} which listens for incoming keys.
    */
   public ContentPane(Frame frame, SwingKeyboardReader keyReader)
   {
      this.keyReader = keyReader;
      this.frame     = frame;

      if (keyReader == null)
      {
         throw new NullPointerException("The key reader can not be null!");
      }
      
      // listen for keystrokes
      frame.addKeyListener(keyReader);
      
      // disable focus processing on TAB and SHIFT-TAB, otherwise those
      // keys are not ever available for consumption
      frame.setFocusTraversalKeys(KeyboardFocusManager.FORWARD_TRAVERSAL_KEYS,
                                  new HashSet<AWTKeyStroke>());
      frame.setFocusTraversalKeys(KeyboardFocusManager.BACKWARD_TRAVERSAL_KEYS,
                                  new HashSet<AWTKeyStroke>());
   }

   /**
    * Capture and return the terminal screen as an image.
    *
    * @return   An image representing the terminal screen that is being
    *           displayed.
    */
   public Image capture()
   {
      BufferedImage image = new BufferedImage(getWidth(),
                                              getHeight(),
                                              BufferedImage.TYPE_4BYTE_ABGR);
      
      printAll(image.createGraphics());
      
      return image;
   }
   
   /**
    * Create a new printer job, open the job dialog for the user and if
    * printing is not canceled by the user, print the current terminal screen
    * to the chosen printer, with the configured job settings.
    */
   public void print()
   {
      PrinterJob job = PrinterJob.getPrinterJob();
      job.setPrintable(this);
      
      if (job.printDialog())
      {
         try
         {
            job.print();
         }
         
         catch (PrinterException pe)
         {
            // ignore
         }
      }      
   }
   
   /**
    * Low level printing worker that is called back by the Java printer job
    * in order to generate the print output. This will only ever print a single
    * page of output. It does not scale the print output to the page, so it
    * is possible that the output will be cut off to the right and on the
    * bottom.
    *
    * @param    g
    *           The graphics instance into which to draw.
    * @param    pf
    *           Details of the printer page being processed.
    * @param    page
    *           The page number being requested. Only the 0 page will ever
    *           be honored.
    *
    * @return   <code>PAGE_EXISTS</code> if the page was printed or
    *           <code>NO_SUCH_PAGE</code> if any page number other than 0
    *           is requested.
    */
   public int print(Graphics g, PageFormat pf, int page)
   throws PrinterException
   {
      if (page > 0)
      {
         return NO_SUCH_PAGE;
      }

      Graphics2D graphics = (Graphics2D) g;
      graphics.translate(pf.getImageableX(), pf.getImageableY());

      printAll(graphics);

      return PAGE_EXISTS;
   }   
   
   /**
    * Cause the outer frame to be disposed, exiting the application or applet.
    * This is important since <code>System.exit()</code> cannot be used for
    * applets.
    */
   public void quit()
   {
      // protect against multiple calls in the same session
      if (frame.isDisplayable())
         frame.dispose();
   }
   
   /**
    * Read the next key from the FIFO key input buffer.  This will block until
    * a key is available.
    *
    * @return   The next key in the buffer.
    */
   public int readKey()
   {
      return keyReader.read();
   }
   
   /**
    * Insert the given characters into the keyboard input buffer (at the end, in FIFO order).
    *
    * @param    data
    *           Characters to insert into the "keyboard" buffer.
    */
   public void paste(String data)
   {
      // TODO: this won't work with interactive child processes
      keyReader.insert(data);
   }
   
   /**
    * Switch between P2J client and interactive client.
    * 
    * @param   toVT100
    *          If <code>true</code> show interactive process screen otherwise P2J client screen.
    */
   public void swapKeyboards(boolean toVT100)
   {
      JFrame app = (JFrame) frame;
      
      KeyListener[] list = app.getKeyListeners();
      
      // if called multiple times with the same "direction" flag, the same listener would
      // be added multiple times, instead we just clear the entire list
      for (KeyListener kl : list)
      {
         app.removeKeyListener(kl);
      }
            
      if (toVT100)
      {
         app.addKeyListener((KeyListener) keyboardVT100);
      }
      else
      {
         app.addKeyListener(keyReader);
      }
   }

   /**
    * Set keyboard listener for VT100 emulator.
    *  
    * @param   keyboard
    *          A VT100 keyboard instance.
    */
   public void setVT100Keyboard(VT100ProgressKeyboard keyboard)
   {
      keyboardVT100 = keyboard;
   }
}
