/*
** Module   : PopupTrigger.java
** Abstract : helper to support popup menu mouse event processing
**
** Copyright (c) 2011-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 GES 20110921 Created initial version.
** 002 GES 20150506 Moved to a driver package.
*/

package com.goldencode.p2j.ui.client.driver.swing;

import java.awt.event.*;
import javax.swing.*;

/**
 * Helper to support popup menu mouse event processing. This class provides
 * a generic mouse event handler that detects when a mouse event is a popup
 * trigger and in that case it will cause the menu to appear.
 * <p>
 * The event type that is a popup trigger will vary on different platforms.
 * This helper class relies upon <code>MouseEvent.isPopupTrigger</code> to
 * detect when the popup menu should be displayed. 
 */
public class PopupTrigger
extends MouseAdapter
{
   /** The menu to be displayed. */
   private JPopupMenu popup = null;
   
   /**
    * Construct an instance.
    *
    * @param    popup
    *           The menu to be displayed when the popup trigger event occurs.
    */
   public PopupTrigger(JPopupMenu popup)
   {
      this.popup = popup;
   }
   
   /**
    * Process the event, causing the associated popup menu to be displayed
    * if the given event is a popup trigger event.
    *
    * @param    evt
    *           The event to process.
    */
   public void mousePressed(MouseEvent evt)
   {
      handlePopup(evt);
   }

   /**
    * Process the event, causing the associated popup menu to be displayed
    * if the given event is a popup trigger event.
    *
    * @param    evt
    *           The event to process.
    */
   public void mouseReleased(MouseEvent evt)
   {
      handlePopup(evt);
   }

   /**
    * Low level worker to implement event processing, causing the associated
    * popup menu to be displayed if the given event is a popup trigger event.
    *
    * @param    evt
    *           The event to process.
    */
   private void handlePopup(MouseEvent evt)
   {
      if (evt.isPopupTrigger())
      {
         popup.show(evt.getComponent(), evt.getX(), evt.getY());
      }
   }
}

