/*
** Module   : SwingKeyboardReader.java
** Abstract : process swing key events and convert them to 4GL keys
**
** Copyright (c) 2011-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 GES 20110921 Created initial version.
** 002 GES 20111012 Made some static data final and added a comment.
** 003 CA  20141029 Refactored into Linux/Windows specific implementations.
** 004 GES 20150506 Moved to a driver package.
*/

package com.goldencode.p2j.ui.client.driver.swing;

import java.awt.event.*;
import java.util.*;
import java.util.concurrent.*;

// selectively import because this is more swing than p2j
import com.goldencode.p2j.ui.client.event.Key;

/**
 * This class processes Swing key events and converts them into 4GL keys. The
 * 7-bit ASCII range is generally handled using {@link #keyTyped} and all other
 * key codes are handled by {@link #keyPressed}.  {@link #keyReleased} is not
 * used since Progress generates all key input on the key press instead of on
 * the key releasse.
 * <p>
 * All input that is not dropped is translated as needed and then is stored
 * in a FIFO queue which represents the key input buffer for the application.
 * This key input is read using {@link #read}.
 * <p>
 * Known Issues and/or TODOs
 * <p>
 * <ul>
 *   <li> refactor this code into CHUI-specific and generic Swing code (this
 *        generic code would be usable by a GUI implementation)
 *   <li> confirm that the numeric keypad (i.e. numpad) works the same on a
 *        keyboard with NUMLOCK off
 * </ul> 
 */
public abstract class SwingKeyboardReader
extends KeyAdapter
{
   // these maps are read-only after initialization so it is OK for them to
   // be static
   
   /** Translation map of Java virtual keys to Progress key codes. */
   protected final Map<Integer, Integer> xlate = new HashMap<>();
   
   /** Incoming keystroke buffer. */
   protected final LinkedBlockingQueue<Integer> queue = new LinkedBlockingQueue<>();
   
   /**
    * Default c'tor.  Populates the key {@link #xlate translations}. 
    */
   public SwingKeyboardReader()
   {
      // The following do not need any translation since the keyTyped() method
      // naturally generates these values:
      //    VK_BACKSPACE
      //    VK_CTRL_*
      //    VK_DELETE
      //    VK_ESCAPE (this is never returned to Progress code anyway)
      //    VK_TAB (the containing frame's focus traversal keys must be
      //            forced to the empty set in order for this to work)
      
      xlate.put(KeyEvent.VK_DOWN, Key.VK_CURSOR_DOWN);
      xlate.put(KeyEvent.VK_END, Key.VK_END);
      xlate.put(KeyEvent.VK_ENTER, Key.VK_ENTER); // Java (10), 4GL (13)
      xlate.put(KeyEvent.VK_F1, Key.VK_F1);
      xlate.put(KeyEvent.VK_F2, Key.VK_F2);
      xlate.put(KeyEvent.VK_F3, Key.VK_F3);
      xlate.put(KeyEvent.VK_F4, Key.VK_F4);
      xlate.put(KeyEvent.VK_F5, Key.VK_F5);
      xlate.put(KeyEvent.VK_F6, Key.VK_F6);
      xlate.put(KeyEvent.VK_F7, Key.VK_F7);
      xlate.put(KeyEvent.VK_F8, Key.VK_F8);
      xlate.put(KeyEvent.VK_F9, Key.VK_F9);
      xlate.put(KeyEvent.VK_F10, Key.VK_F10);
      xlate.put(KeyEvent.VK_F11, Key.VK_F11);
      xlate.put(KeyEvent.VK_F12, Key.VK_F12);
      xlate.put(KeyEvent.VK_F13, Key.VK_F13);
      xlate.put(KeyEvent.VK_F14, Key.VK_F14);
      xlate.put(KeyEvent.VK_F15, Key.VK_F15);
      xlate.put(KeyEvent.VK_F16, Key.VK_F16);
      xlate.put(KeyEvent.VK_F17, Key.VK_F17);
      xlate.put(KeyEvent.VK_F18, Key.VK_F18);
      xlate.put(KeyEvent.VK_F19, Key.VK_F19);
      xlate.put(KeyEvent.VK_F20, Key.VK_F20);
      xlate.put(KeyEvent.VK_HOME, Key.VK_HOME);
      xlate.put(KeyEvent.VK_INSERT, Key.VK_INSERT);
      xlate.put(KeyEvent.VK_LEFT, Key.VK_CURSOR_LEFT);
      xlate.put(KeyEvent.VK_PAGE_DOWN, Key.VK_PAGE_DOWN);
      xlate.put(KeyEvent.VK_PAGE_UP, Key.VK_PAGE_UP);
      xlate.put(KeyEvent.VK_RIGHT, Key.VK_CURSOR_RIGHT);
      xlate.put(KeyEvent.VK_TAB, Key.VK_TAB);  // not used for VK_TAB, instead  
                                               // used with SHIFT = VK_BACKTAB
      xlate.put(KeyEvent.VK_UP, Key.VK_CURSOR_UP);
   }
   
   /**
    * Read a keystroke from the key input buffer. This method will block until
    * a key is available.
    *
    * @return   A 4GL-compatible key code matching the input combination which
    *           the user pressed on the keyboard.
    */
   public int read()
   {
      int key = -1;
      
      try
      {
         // use of the queue is thread-safe
         key = queue.take();
      }
      
      catch (InterruptedException ie)
      {
         // ignore (TODO: is this right?)
      }
      
      return key; 
   }
   
   /**
    * Add each character of the string to end of the key input buffer in FIFO
    * order (left to right). Each character is converted directly to an
    * integer during the addition. This is fine for the 7-bit ASCII range, but
    * it may be incorrect for other UNICODE characters.
    *
    * @param    data
    *           The "keys" to insert.
    */
   public void insert(String data)
   {
      // use of the queue is thread-safe
      for (int i = 0; i < data.length(); i++)
      {
         queue.add(new Integer((int) data.charAt(i)));
      }
   }

   /**
    * This method is called when a pressed key is released. This is not used
    * since Progress processes key events on the key press, not on the key
    * release.  In addition, no state needs to be cleaned up using this method
    * so it is a NO-OPERATION.
    *
    * @param    evt
    *           The key event.
    */
   public synchronized void keyReleased(KeyEvent evt)
   {
   }
}

