/*
** Module   : WatchdogTimer.java
** Abstract : a watchdog timer implementation
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- ----------------------Description--------------------------
** 001 MAG 20140131 Created initial version.
** 002 MAG 20140213 Add log messages.
** 003 MAG 20140213 Fixed loop exit condition.
** 004 MAG 20140220 Add join statement on exit.
** 005 GES 20150308 Moved to a more generic package location.
*/

package com.goldencode.p2j.ui.client.driver.web;

/**
 * A watchdog timer thread implementation.  Once started, if the timer thread is not stopped
 * before the timer expires then the JVM process is ended using <code>System.exit()</code>. This
 * is useful to cleanup a process which is waiting for a connection that never arrives.
 * <p>
 * The thread should only be started when the process is in a state where it should be exited
 * at the end of the timeout. 
 * <p>
 * IMPORTANT: Do not use logging anywhere inside this class. When an authentication plugin
 * is running during session establishment with the P2J server using logging no messages are 
 * written into the log and the application becomes deadlocked.
 */
public class WatchdogTimer
extends Thread
{
   /** Time to wait for thread exit. */
   private static final long WAIT_TO_DIE = 5000;

   /** One minute in milliseconds. */
   private static final long ONE_MINUTE = 60000;
   
   /** System time when started. */
   private volatile long startTime;

   /** Time out value in milliseconds. */
   private volatile long timeout;
   
   /** Flag to control thread abort. */
   private volatile boolean running = true;
   
   /**
    * Create a new instance and start the timer thread.
    * 
    * @param    timeout
    *           The time out value in milliseconds.
    */
   public WatchdogTimer(long timeout)
   {
      this.setName("watchdog");
      this.timeout = timeout;
      this.start();
   }

   /**
    * Because the Thread.sleep time is not guaranteed to be accurate, the elapsed time is
    * calculated once per minute. If the elapsed time is greater than the timeout value, a 
    * JVM process exit is forced (using <code>System.exit()</code>).  Any spurious wake-ups
    * are avoided using a loop condition.
    */
   @Override
   public void run()
   {
      // start time
      startTime = System.currentTimeMillis();

      while (running)
      {
         try
         {
            Thread.sleep(WatchdogTimer.ONE_MINUTE);
         }
         catch (InterruptedException e)
         {
            e.printStackTrace();
         }
         
         // every minute check elapsed time and force an exit on timeout
         long elapsedTime = System.currentTimeMillis() - startTime;
         
         if (running && elapsedTime >= timeout)
         {
            // timer has popped, process goes down now
            System.exit(-1);
         }         
      }
   }
   
   /**
    * Terminate the watchdog timer thread.
    */
   public void kill()
   {
      running = false;
      this.interrupt();
      
      // wait for thread to exit
      try
      {
         this.join(WatchdogTimer.WAIT_TO_DIE);
      }
      catch (InterruptedException e)
      {
         e.printStackTrace();
      }
   }   
}
