/*
** Module   : PopupMenu.java
** Abstract : A popup menu displayed via mouse and associated with a widget.
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140805 Created initial version.
** 002 MAG 20140902 Use GUI driver system color API to get default colors for widgets drawing.
** 003 HC  20140911 GUI coordinate support - data types refactoring.
** 004 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 005 CA  20141021 Improved the drawing and behaviour of the popup menu. 
** 006 MAG 20141029 Added clipping rectangle when drawing. 
** 007 HC  20141021 Implemented window move/position support. 
** 008 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 009 HC  20150323 Removed business logic from config classes.
** 010 GES 20150507 Added a comment to clarify that AWT events usage is intentional.
*/

package com.goldencode.p2j.ui.client.gui;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.MouseEvent;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Implementation of a popup menu.
 */
public class PopupMenu
extends AbstractContainer<GuiOutputManager>
implements WidgetWithConfig<BaseConfig>
{
   /** The widget configuration. */
   private BaseConfig config;

   /** The graphics driver. */
   private GuiDriver gd;

   /** Flag indicating that a popup option is being drawn. */
   private boolean processingOption = false;
   
   /**
    * Basic c'tor.
    * 
    * @param    id
    *           The widget's ID.
    */
   public PopupMenu(WidgetId id)
   {
      this.gd = (GuiDriver) OutputManager.getDriver();
      
      this.config = ConfigManager.getInstance().resolveWidgetConfig(id, BaseConfig.class);
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, BaseConfig cfg)
   {
      // no-op
   }

   /**
    * Get the widget config.
    * 
    * @return   See above.
    */
   @SuppressWarnings("unchecked")
   @Override
   public BaseConfig config()
   {
      return config;
   }
   
   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    */
   @Override
   public void afterConfigUpdate(BaseConfig beforeUpdate)
   {
      super.afterConfigUpdateBase(beforeUpdate);
   }

   /**
    * Draw this popup menu.
    */
   @Override
   public void draw()
   {
      final NativePoint origin = physicalLocation();
      final NativeDimension d = physicalDimension();
      final NativeRectangle clip = clipRectangle(origin, d, true);
      
      gd.draw(origin, clip, new Runnable()
      {
         @Override
         public void run()
         {
            gd.setColor(gd.getSysColor(SysColor.COLOR_MENU));
            gd.fill3DRect(0, 0, d.width, d.height, true);
            
            PopupMenu.super.draw();
         }
      });
   }
   
   /**
    * Lay out the options of this popup.
    */
   @Override
   public void doLayout()
   {
      Widget<GuiOutputManager>[] widgets = widgets();
      CoordinatesConversion c = window().screen().coordinates();
      
      double x = 0;
      double y = 0;
      double width = 0;
      
      for (int i = 0; i < widgets.length; i++)
      {
         PopupOption w = (PopupOption) widgets[i];
         
         w.setLocation(x, y);
         
         int ny = c.rowToNative(y);
         int nx = c.columnToNative(x);
         w.setPhysicalLocation(nx, ny);
         
         double h = w.height();
         y += h;

         width = Math.max(width, w.width());
      }
      
      setHeight(y);
      setWidth(width);
   }

   /**
    * Add a new option to this popup.
    * 
    * @param    label
    *           The option's label.
    * @param    action
    *           The option's action, when clicked.
    * 
    * @return   The created option.
    */
   public PopupOption addOption(String label, Runnable action)
   {
      PopupOption option = new PopupOption(WidgetId.nextID(), label, action);
      this.add(option);
      
      return option;
   }
   
   /**
    * Add a new option to this popup.
    * 
    * @param    active
    *           Flag indicating if the option is active or not.
    * @param    label
    *           The option's label.
    * @param    action
    *           The option's action, when clicked.
    */
   public void addOption(boolean active, String label, Runnable action)
   {
      this.add(new PopupOption(WidgetId.nextID(), active, label, action));
   }
   
   /**
    * Hide/show popup menu and all its options.
    * 
    * @param   visible
    *          New visible state.
    */
   @Override
   public void setVisible(boolean visible)
   {
      super.setVisible(visible);
      
      for (Widget<GuiOutputManager> opt : widgets())
      {
         opt.setVisible(visible);
      }
   }
   
   /**
    * Notification of a mouse clicked event occurred for this widget.
    * <p>
    * If no popup option is currently being processed, the popup-menu will be hidden.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      if (!processingOption)
      {
         hide();
      }
   }
   
   /**
    * Notification of a mouse pressed event occurred for this widget.
    * <p>
    * If no popup option is currently being processed, the popup-menu will be hidden.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      if (!processingOption)
      {
         hide();
      }
   }
   
   /**
    * Hide widget.
    */
   @Override
   public void hide()
   {
      gd.deregisterWidget(config().id.asInt());
      super.hide();
   }
   
   /**
    * Check if the widget implementation can raise legacy, 4GL-style, mouse events.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   public boolean raiseLegacyMouseEvents()
   {
      return false;
   }

   /**
    * Change the state of the {@link #processingOption} flag.
    * 
    * @param    flag
    *           The new state; <code>true</code> means a popup option (for this menu) is currenlty
    *           being processed.
    */
   public void processingOption(boolean flag)
   {
      this.processingOption = flag;
   }
}
